#ifndef GL_H
#define GL_H

#include "fix.h"

//Either colors or textures, but not both
#define TEXTURE_SUPPORT

//Works best with 300
#define NEAR_PLANE 300

#define MATRIX_STACK_SIZE 5

//Change if you want to, but it may not work.
typedef Fix<9, int_fast32_t> GLFix;

/* Column vectors and matrices:
 * [ [0][0] [0][1] [0][2] [0][3] ]   [x]
 * [ [1][0] [1][1] [1][2] [1][3] ] * [y]
 * [ [2][0] [2][1] [2][2] [2][3] ]   [z]
 * [ [3][0] [3][1] [3][2] [3][3] ]   [1]
 */

typedef struct VERTEX { GLFix x; GLFix y; GLFix z; GLFix u; GLFix v; } VERTEX;
typedef uint16_t COLOR;

typedef struct TEXTURE { int width; int height; COLOR *bitmap; } TEXTURE;

class MATRIX {
public:
    MATRIX() { memset(data, 0, sizeof(data)); }
    GLFix data[4][4];
};

#define GL_COLOR_BUFFER_BIT 1<<0
#define GL_DEPTH_BUFFER_BIT 1<<1

enum GLDrawMode
{
    GL_TRIANGLES,
    GL_QUADS,
    GL_QUAD_STRIP //Not really tested
};

void set_screen(uint16_t* screenBuf);
void init_gl();
void uninit_gl();
void set_color(COLOR c);
void rotate_x(GLFix a);
void rotate_y(GLFix a);
void rotate_z(GLFix a);
void add_vertex(VERTEX* vertex);

void glLoadIdentity();
void glBegin(GLDrawMode mode);
inline void glEnd() { }
void glClear(int buffers);
void glTranslatef(GLFix x, GLFix y, GLFix z);
void glBindTexture(TEXTURE *tex);
void glTexCoord2f(GLFix nu, GLFix nv);
void glColor3f(GLFix r, GLFix g, GLFix b);
void glVertex3f(GLFix x, GLFix y, GLFix z);
void glScale3f(GLFix x, GLFix y, GLFix z);
void glPushMatrix();
void glPopMatrix();

#endif
