extern "C" {
#include <common.h>
}

#include "gl.h"
#include "fastmath.h"

#include "textures.h"

#define HEIGHT 240
#define WIDTH 320

#define FIELD_WIDTH 25
#define FIELD_HEIGHT 19
#define SEGMENT_SIZE GLFix(70)

typedef struct snake_segment {
    int x;
    int y;
    struct snake_segment *next;
    struct snake_segment *last;
} snake_segment;

enum DIRECTION {
    UP = 180,
    DOWN = 0,
    LEFT = 270,
    RIGHT = 90
};

void operator delete(void *p)
{
    free(p);
}

static snake_segment *head = nullptr;
static DIRECTION direction = RIGHT;
static int food_x = 0, food_y = 0;
static int length = 0;
static int t = 0;

static void render_cube(GLFix x, GLFix y, GLFix z, GLFix w, GLFix h, GLFix l, GLFix u1, GLFix u2, GLFix v1, GLFix v2)
{
    w += x;
    h += y;
    l += z;

    glBegin(GL_QUADS);

    //if(draw_front)
    {
        glTexCoord2f(u1, v2);
        glVertex3f(x, y, z);
        glTexCoord2f(u1, v1);
        glVertex3f(x, h, z);
        glTexCoord2f(u2, v1);
        glVertex3f(w, h, z);
        glTexCoord2f(u2, v2);
        glVertex3f(w, y, z);
    }

    //if(draw_back)
    {
        glTexCoord2f(u2, v2);
        glVertex3f(w, y, l);
        glTexCoord2f(u2, v1);
        glVertex3f(w, h, l);
        glTexCoord2f(u1, v1);
        glVertex3f(x, h, l);
        glTexCoord2f(u1, v2);
        glVertex3f(x, y, l);
    }

    //if(draw_left)
    {
        glTexCoord2f(u1, v2);
        glVertex3f(x, y, l);
        glTexCoord2f(u1, v1);
        glVertex3f(x, h, l);
        glTexCoord2f(u2, v1);
        glVertex3f(x, h, z);
        glTexCoord2f(u2, v2);
        glVertex3f(x, y, z);
    }

    //if(draw_right)
    {
        glTexCoord2f(u1, v2);
        glVertex3f(w, y, z);
        glTexCoord2f(u1, v1);
        glVertex3f(w, h, z);
        glTexCoord2f(u2, v1);
        glVertex3f(w, h, l);
        glTexCoord2f(u2, v2);
        glVertex3f(w, y, l);
    }

    //if(draw_top)
    {
        glTexCoord2f(u1, v2);
        glVertex3f(x, h, z);
        glTexCoord2f(u1, v1);
        glVertex3f(x, h, l);
        glTexCoord2f(u2, v1);
        glVertex3f(w, h, l);
        glTexCoord2f(u2, v2);
        glVertex3f(w, h, z);
    }

    //if(draw_bottom)
    {
        glTexCoord2f(u2, v1);
        glVertex3f(w, y, z);
        glTexCoord2f(u2, v2);
        glVertex3f(w, y, l);
        glTexCoord2f(u1, v2);
        glVertex3f(x, y, l);
        glTexCoord2f(u1, v1);
        glVertex3f(x, y, z);
    }
    glEnd();
}

void render_body(snake_segment *seg)
{
    glBindTexture(&snakebody);
    glPushMatrix();
    glTranslatef(SEGMENT_SIZE * seg->x, 0, SEGMENT_SIZE * seg->y);
    render_cube(0, 0, 0, SEGMENT_SIZE, SEGMENT_SIZE, SEGMENT_SIZE, 0, 1, 0, 1);
    glPopMatrix();
}

void render_head()
{
    GLFix x = 0, y = 0, z = 0;
    GLFix h = SEGMENT_SIZE, l = SEGMENT_SIZE, w = SEGMENT_SIZE;

    glPushMatrix();

    glTranslatef(SEGMENT_SIZE * head->x, 0, SEGMENT_SIZE * head->y);
    glTranslatef(SEGMENT_SIZE/2, SEGMENT_SIZE/2, SEGMENT_SIZE/2);
    rotate_y(direction);
    glTranslatef(-SEGMENT_SIZE/2, -SEGMENT_SIZE/2, -SEGMENT_SIZE/2);

    glBegin(GL_QUADS);

    glBindTexture(&snakeface);
    glTexCoord2f(0, 1);
    glVertex3f(x, y, z);
    glTexCoord2f(0, 0);
    glVertex3f(x, h, z);
    glTexCoord2f(1, 0);
    glVertex3f(w, h, z);
    glTexCoord2f(1, 1);
    glVertex3f(w, y, z);

    glBindTexture(&snakebody);
    glTexCoord2f(1, 1);
    glVertex3f(w, y, l);
    glTexCoord2f(1, 0);
    glVertex3f(w, h, l);
    glTexCoord2f(0, 0);
    glVertex3f(x, h, l);
    glTexCoord2f(0, 1);
    glVertex3f(x, y, l);

    glBindTexture(&snakehead);
    glTexCoord2f(1, 1);
    glVertex3f(x, y, l);
    glTexCoord2f(1, 0);
    glVertex3f(x, h, l);
    glTexCoord2f(0, 0);
    glVertex3f(x, h, z);
    glTexCoord2f(0, 1);
    glVertex3f(x, y, z);

    glTexCoord2f(0, 1);
    glVertex3f(w, y, z);
    glTexCoord2f(0, 0);
    glVertex3f(w, h, z);
    glTexCoord2f(1, 0);
    glVertex3f(w, h, l);
    glTexCoord2f(1, 1);
    glVertex3f(w, y, l);

    glBindTexture(&snakebody);
    glTexCoord2f(0, 1);
    glVertex3f(x, h, z);
    glTexCoord2f(0, 0);
    glVertex3f(x, h, l);
    glTexCoord2f(1, 0);
    glVertex3f(w, h, l);
    glTexCoord2f(1, 1);
    glVertex3f(w, h, z);

    glTexCoord2f(1, 0);
    glVertex3f(w, y, z);
    glTexCoord2f(1, 1);
    glVertex3f(w, y, l);
    glTexCoord2f(0, 1);
    glVertex3f(x, y, l);
    glTexCoord2f(0, 0);
    glVertex3f(x, y, z);

    glEnd();

    glPopMatrix();
}

void render_apple(GLFix x, GLFix y)
{
    glBindTexture(&apple);

    render_cube(x * SEGMENT_SIZE, 0, y * SEGMENT_SIZE, SEGMENT_SIZE, SEGMENT_SIZE, SEGMENT_SIZE, 0, 1, 0, 1);
}

void draw_wall(GLFix x, GLFix y)
{
    glBindTexture(&wall);

    render_cube(x * SEGMENT_SIZE, 0, y * SEGMENT_SIZE, SEGMENT_SIZE, SEGMENT_SIZE, SEGMENT_SIZE, 0, 1, 0, 1);
}

bool on_snake(int x, int y)
{
    snake_segment *current = head;
    while(current)
    {
        if(x == current->x && y == current->y)
            return true;

        current = current->next;
    }

    return false;
}

void spawn_food()
{
    do
    {
        int x = rand();
        int y = rand();
        food_x = x % FIELD_WIDTH;
        food_y = y % FIELD_HEIGHT;
    }
    while(on_snake(food_x, food_y));
}

void reset_game()
{
    if(head != nullptr)
    {
        struct snake_segment *current = head, *next;
        while(current)
        {
            next = current->next;
            delete current;
            current = next;
        }
    }

    length = 3;
    head = new snake_segment;
    head->last = nullptr;
    head->next = new snake_segment;
    head->next->last = head;
    head->next->next = new snake_segment;
    head->next->next->next = nullptr;
    head->next->next->last = head->next;

    head->y = head->next->y = head->next->next->y = 10;
    head->x = 12;
    head->next->x = 11;
    head->next->next->x = 10;

    direction = RIGHT;

    spawn_food();
}

int main()
{
    if(!lcd_isincolor())
    {
        show_msgbox("Error", "Your LCD hasn't got enough colors :-P");
        return EXIT_FAILURE;
    }

    uint16_t** screen_address = (uint16_t**)0xC0000010;
    uint16_t* real_screen = *screen_address;
    uint16_t *screen_buffer = new uint16_t[WIDTH*HEIGHT];

    srand(*reinterpret_cast<unsigned int*>(screen_buffer));
    rand();

    init_gl();

    set_screen(screen_buffer);

    reset_game();

    GLFix x = SEGMENT_SIZE*FIELD_WIDTH/2 - SEGMENT_SIZE*2, y = 1800, z = SEGMENT_SIZE*FIELD_HEIGHT/2 - SEGMENT_SIZE*2;
    GLFix xr = 90, yr = 0;
    GLFix incr = 1;

    while(!isKeyPressed(KEY_NSPIRE_ESC))
    {
        glColor3f(1, 1, 1); //White background
        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

        glPushMatrix();

        rotate_x(GLFix(360) - xr.normalise_angle());
        //rotate_y(GLFix(360) - yr.normalise_angle());
        glTranslatef(-x, -y, -z);

        if(isKeyPressed(KEY_NSPIRE_W))
            z += incr;
        if(isKeyPressed(KEY_NSPIRE_S))
            z -= incr;
        if(isKeyPressed(KEY_NSPIRE_D))
            x += incr;
        if(isKeyPressed(KEY_NSPIRE_A))
            x -= incr;
        if(isKeyPressed(KEY_NSPIRE_SHIFT))
            y -= incr;
        if(isKeyPressed(KEY_NSPIRE_SPACE))
            y += incr;
        if(isKeyPressed(KEY_NSPIRE_LEFT))
            yr -= incr;
        if(isKeyPressed(KEY_NSPIRE_RIGHT))
            yr += incr;
        if(isKeyPressed(KEY_NSPIRE_UP))
            xr -= incr;
        if(isKeyPressed(KEY_NSPIRE_DOWN))
            xr += incr;

        if(isKeyPressed(KEY_NSPIRE_8) && direction != DOWN)
            direction = UP;
        else if(isKeyPressed(KEY_NSPIRE_2) && direction != UP)
            direction = DOWN;
        else if(isKeyPressed(KEY_NSPIRE_6) && direction != LEFT)
            direction = RIGHT;
        else if(isKeyPressed(KEY_NSPIRE_4) && direction != RIGHT)
            direction = LEFT;
        else if(isKeyPressed(KEY_NSPIRE_ESC))
            break;

        render_head();

        snake_segment *body = head->next;

        while(body)
        {
            render_body(body);
            body = body->next;
        }

        render_apple(food_x, food_y);

        for(int y = 0; y < FIELD_HEIGHT; y++)
        {
            draw_wall(-1, y);
            draw_wall(FIELD_WIDTH, y);
        }

        for(int x = -1; x <= FIELD_WIDTH; x++)
        {
            draw_wall(x, -1);
            draw_wall(x, FIELD_HEIGHT);
        }

        glPopMatrix();

        memcpy(real_screen, screen_buffer, 320*240*2);

        //Game logic
        t++;
        if(t != 3)
            continue;
        else
            t = 0;

        snake_segment *new_head;

        if(head->x == food_x && head->y == food_y)
        {
            new_head = new snake_segment;
            spawn_food();
            length++;
        }
        else
        {
            body = head;
            while(1)
            {
                if(body->next == nullptr)
                {
                    new_head = body;
                    new_head->last->next = nullptr;
                    break;
                }
                else
                    body = body->next;
            }
        }

        new_head->last = nullptr;
        new_head->next = head;
        head->last = new_head;
        new_head->x = head->x;
        new_head->y = head->y;

        switch(direction)
        {
        case UP:
            new_head->y = head->y + 1;
            break;
        case DOWN:
            new_head->y = head->y - 1;
            break;
        case RIGHT:
            new_head->x = head->x + 1;
            break;
        case LEFT:
            new_head->x = head->x - 1;
            break;
        }

        if(new_head->x == -1 || new_head->x == FIELD_WIDTH || new_head->y == -1 || new_head->y == FIELD_HEIGHT || on_snake(new_head->x, new_head->y))
        {
            head = new_head;
            reset_game();
            continue;
        }

        head = new_head;
    }

    uninit_gl();

    *screen_address = real_screen;
    return EXIT_SUCCESS;
}
